@@ -3,14 +3,14 @@ package MojoX::Redis;
 use strict;
 use warnings;
 
-our $VERSION = 0.7;
+our $VERSION = 0.8;
 use base 'Mojo::Base';
 
 use Mojo::IOLoop;
 use List::Util   ();
 use Mojo::Util   ();
 use Scalar::Util ();
-use Carp 'croak';
+require Carp;
 
 __PACKAGE__->attr(server   => '127.0.0.1:6379');
 __PACKAGE__->attr(ioloop   => sub { Mojo::IOLoop->singleton });
@@ -26,6 +26,13 @@ __PACKAGE__->attr(
     }
 );
 
+__PACKAGE__->attr(
+    protocol_redis => sub {
+        require Protocol::Redis;
+        "Protocol::Redis";
+    }
+);
+
 our @COMMANDS = qw/
   append auth bgrewriteaof bgsave blpop brpop brpoplpush config_get config_set
   config_resetstat dbsize debug_object debug_segfault decr decrby del discard
@@ -40,13 +47,13 @@ our @COMMANDS = qw/
   unsubscribe unwatch watch zadd zcard zcount zincrby zinterstore zrange
   zrangebyscore zrank zrem zremrangebyrank zremrangebyscore zrevrange
   zrevrangebyscore zrevrank zscore zunionstore
-/;
+  /;
 
 sub AUTOLOAD {
     my ($package, $cmd) = our $AUTOLOAD =~ /^([\w\:]+)\:\:(\w+)$/;
 
     Carp::croak(qq|Can't locate object method "$cmd" via "$package"|)
-        unless List::Util::first {$_ eq $cmd} @COMMANDS;
+      unless List::Util::first { $_ eq $cmd } @COMMANDS;
 
     my $self = shift;
 
@@ -54,7 +61,8 @@ sub AUTOLOAD {
     my $cb   = $args->[-1];
     if (ref $cb ne 'CODE') {
         $cb = undef;
-    } else {
+    }
+    else {
         pop @$args;
     }
 
@@ -86,15 +94,17 @@ sub connect {
 
     Scalar::Util::weaken $self;
 
+    $self->{_protocol} = $self->_create_protocol;
+
     # connect
     $self->{_connecting} = 1;
     $self->{_connection} = $self->ioloop->connect(
         {   address    => $address,
             port       => $port,
             on_connect => sub { $self->_on_connect(@_) },
-            on_read    => sub { $self->_read_wait_command(@_) },
+            on_read    => sub { $self->_on_read(@_) },
             on_error   => sub { $self->_on_error(@_) },
-            on_hup     => sub { $self->_on_hup(@_) },
+            on_close   => sub { $self->_on_hup(@_) },
         }
     );
 
@@ -120,18 +130,11 @@ sub execute {
 
     unshift @$args, uc $command;
 
-    my $message = '*' . scalar(@$args) . "\r\n";
-    foreach my $token (@$args) {
-        Mojo::Util::encode($self->encoding, $token) if $self->encoding;
-        $message .= '$' . length($token) . "\r\n" . "$token\r\n";
-    }
-    $message .= "\r\n";
-
     my $mqueue = $self->{_message_queue} ||= [];
     my $cqueue = $self->{_cb_queue}      ||= [];
 
 
-    push @$mqueue, $message;
+    push @$mqueue, $args;
     push @$cqueue, $cb;
 
     $self->connect unless $self->{_connection};
@@ -154,11 +157,37 @@ sub stop {
     return $self;
 }
 
+sub _create_protocol {
+    my $self = shift;
+
+    my $protocol = $self->protocol_redis->new(
+        api        => 1,
+        on_message => sub {
+            my ($parser, $command) = @_;
+            $self->_return_command_data($command);
+        }
+    );
+
+    Carp::croak(q/Protocol::Redis implementation doesn't support APIv1/)
+      unless $protocol;
+
+    $protocol;
+}
+
 sub _send_next_message {
     my ($self) = @_;
 
     if ((my $c = $self->{_connection}) && !$self->{_connecting}) {
-        while (my $message = shift @{$self->{_message_queue}}) {
+        while (my $args = shift @{$self->{_message_queue}}) {
+            my $cmd_arg = [];
+            my $cmd = {type => '*', data => $cmd_arg};
+            foreach my $token (@$args) {
+                Mojo::Util::encode($self->encoding, $token)
+                  if $self->encoding;
+                push @$cmd_arg, {type => '$', data => $token};
+            }
+            my $message = $self->{_protocol}->encode($cmd);
+
             $self->ioloop->write($c, $message);
         }
     }
@@ -173,8 +202,32 @@ sub _on_connect {
     $self->_send_next_message;
 }
 
+sub _reencode_message {
+    my ($self, $message) = @_;
+
+    my ($type, $data) = @{$message}{'type', 'data'};
+    if ($type eq '-') {
+        $self->error($data);
+        $self->on_error->($self);
+        return;
+    }
+    elsif ($type ne '*') {
+        return [$data];
+    }
+    else {
+        my $reencoded_data = [];
+        foreach my $item (@$data) {
+            my $message = $self->_reencode_message($item);
+            push @$reencoded_data, $message;
+        }
+        return $reencoded_data;
+    }
+}
+
 sub _return_command_data {
-    my ($self, $data) = @_;
+    my ($self, $message) = @_;
+
+    my $data = $self->_reencode_message($message);
 
     my $cb = shift @{$self->{_cb_queue}};
     if ($cb) {
@@ -223,177 +276,10 @@ sub _inform_queue {
     $self->{_queue} = [];
 }
 
-sub _read_wait_command {
+sub _on_read {
     my ($self, $ioloop, $id, $chunk) = @_;
 
-    Scalar::Util::weaken $self;
-    my $cmd = substr $chunk, 0, 1, '';
-    if (!defined $chunk || $chunk eq '') {
-
-        # Wait next command
-        $ioloop->on_read($id => sub { $self->_read_wait_command(@_) });
-    }
-    elsif (List::Util::first { $cmd eq $_ } ('+', '-', ':')) {
-
-        # Just a simple one line command
-        $self->{_read_cmd_string} = '';
-        if ($cmd ne '-') {
-            $self->{_read_cb} = sub {
-                $self->_return_command_data(shift);
-                $self->_read_wait_command($self->ioloop, $id, shift);
-            };
-        }
-        else {
-            $self->{_read_cb} = sub {
-                $self->error(shift->[0]);
-                $self->on_error->($self);
-                $self->_return_command_data(undef);
-                $self->_read_wait_command($self->ioloop, $id, shift);
-            };
-        }
-
-        $ioloop->on_read($id => sub { $self->_read_string_command(@_); });
-        $self->_read_string_command($self->ioloop, $id, $chunk);
-
-    }
-    elsif ($cmd eq '$') {
-
-        # Bulk command, not a big deal
-        $self->{_read_cb} = sub {
-            $self->_return_command_data(shift);
-            $self->_read_wait_command($self->ioloop, $id, shift);
-        };
-
-        # Yes, it should have leading $
-        $self->_read_bulk_command($ioloop, $id, "\$$chunk");
-    }
-    elsif ($cmd eq '*') {
-        $self->{_read_cb} = sub {
-            $self->_return_command_data(shift);
-            $self->_read_wait_command($self->ioloop, $id, shift);
-        };
-        $self->_read_multi_bulk_command($ioloop, $id, $chunk);
-    }
-    else {
-        die qq{Strange input "$cmd$chunk"};
-    }
-}
-
-sub _read_string_command {
-    my ($self, $ioloop, $id, $chunk) = @_;
-
-    my $str = $self->{_read_cmd_string} .= $chunk;
-    my $i = index $str, "\r\n";
-
-    if ($i >= 0) {
-
-        # Got full command
-        my $result = substr $str, 0, $i, '';
-        substr $str, 0, 2, '';    # Delete \r\n
-
-        #print "## $result\n## $str\n";
-        my $cb = $self->{_read_cb};
-        delete $self->{_read_cb};
-        delete $self->{_read_cmd_string};
-        $cb->([$result], $str);
-    }
-}
-
-sub _read_multi_bulk_command {
-    my ($self, $ioloop, $id, $chunk) = @_;
-
-    delete $self->{_read_cmd_num};
-
-    my $mbulk_cb = $self->{_read_cb};
-
-    my $results = [];
-    my $mbulk_process;
-    $mbulk_process = sub {
-        push @$results, shift->[0];
-
-        if (scalar @$results == $self->{_read_cmd_num}) {
-            $mbulk_process = undef;
-            $mbulk_cb->($results, shift);
-        }
-        else {
-
-            # Read another string
-            #print "### Got another result: ", $results->[-1], "\n";;
-            $self->{_read_cb} = $mbulk_process;
-            $self->_read_bulk_command($ioloop, $id, shift);
-        }
-    };
-
-    # Read number of commands
-    $self->{_read_cb} = sub {
-        $self->{_read_cmd_num} = shift->[0];
-        my $chunk = shift;
-
-        if ($self->{_read_cmd_num} < 1) {
-            $mbulk_process = undef;
-            $mbulk_cb->($results, $chunk);
-            return;
-        }
-
-        #print "Got #commands: ", $self->{_read_cmd_num}, "\n";
-        $self->{_read_cb} = $mbulk_process;
-        $ioloop->on_read($id => sub { $self->_read_bulk_command(@_) });
-        $self->_read_bulk_command($ioloop, $id, $chunk);
-    };
-
-    $ioloop->on_read($id => sub { $self->_read_string_command(@_); });
-    $self->_read_string_command($ioloop, $id, $chunk);
-}
-
-sub _read_bulk_command {
-    my ($self, $ioloop, $id, $chunk) = @_;
-
-    delete $self->{_read_cmd_legth};
-
-    my $bulk_cb = $self->{_read_cb};
-
-    # Read size of string
-    $self->{_read_cb} = sub {
-        my $size  = shift->[0];
-        my $chunk = shift;
-
-        # Delete leading $
-        substr $size, 0, 1, "";
-        $self->{_read_cmd_legth} = $size;
-
-        #print "Got size: ", $self->{_read_cmd_legth}, "\n";
-        $self->{_read_cb} = $bulk_cb;
-
-        if ($size == '-1') {
-            $self->{_read_cb}->([], $chunk);
-        }
-        else {
-            $ioloop->on_read(
-                $id => sub { $self->_read_bulk_command_string(@_) });
-            $self->_read_bulk_command_string($ioloop, $id, $chunk);
-        }
-    };
-
-    $ioloop->on_read($id => sub { $self->_read_string_command(@_); });
-    $self->_read_string_command($ioloop, $id, $chunk);
-}
-
-sub _read_bulk_command_string {
-    my ($self, $ioloop, $id, $chunk) = @_;
-
-    my $str = $self->{_read_cmd_string} .= $chunk;
-
-    #print "String $str\n";
-    if (length $str >= $self->{_read_cmd_legth}) {
-        my $result = substr $str, 0, $self->{_read_cmd_legth}, "";
-        substr $str, 0, 2, "";    # Delete \r\n
-
-        #print "## str result $result\n";
-        my $cb = $self->{_read_cb};
-        delete $self->{_read_cb};
-        delete $self->{_read_cmd_string};
-        $cb->([$result], $str);
-    }
+    $self->{_protocol}->parse($chunk);
 }
 
 1;
@@ -401,7 +287,7 @@ __END__
 
 =head1 NAME
 
-L<MojoX::Redis> - asynchronous Redis client for L<Mojolicious>.
+MojoX::Redis - asynchronous Redis client for L<Mojolicious>.
 
 =head1 SYNOPSIS
 
@@ -473,6 +359,15 @@ dropped, defaults to C<300>.
 
 Encoding used for stored data, defaults to C<UTF-8>.
 
+=head2 C<protocol_redis>
+
+    use Protocol::Redis;
+    $redis->protocol_redis("Protocol::Redis");
+
+L<Protocol::Redis> implementation' constructor for parsing. By default
+L<Protocol::Redis> will be used. Parser library must support
+L<APIv1|Protocol::Redis/APIv1>.
+
 =head1 METHODS
 
 L<MojoX::Redis> supports Redis' methods.
@@ -532,7 +427,7 @@ Starts IOLoop. Shortcut for $redis->ioloop->start;
 
 =head1 SEE ALSO
 
-L<Mojolicious>, L<Mojo::IOLoop>
+L<Protocol::Redis>, L<Mojolicious>, L<Mojo::IOLoop>
 
 =head1 SUPPORT
 
@@ -544,7 +439,7 @@ L<Mojolicious>, L<Mojo::IOLoop>
 
 =head2 Repository
 
-    http://github.com/und3f/mojoliciousx-lexicon
+    https://github.com/und3f/mojox-redis
 
 =head1 AUTHOR
 
@@ -552,7 +447,7 @@ Sergey Zasenko, C<undef@cpan.org>.
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2010, Sergey Zasenko
+Copyright (C) 2010-2011, Sergey Zasenko
 
 This program is free software, you can redistribute it and/or modify it under
 the terms of the Artistic License version 2.0.
